# ../rpg4go/User.py

# =============================================================================
# >> IMPORTS
# =============================================================================
from rpg4go.Level import Level
import rpg4go.config
import rpg4go.skills

# Python
import sqlite3

# Source.Python
from Source import Player
from Source import Engine
from players.entity import get_uniqueid_from_player
from messages import chat


# =============================================================================
# >> CLASS: User
# =============================================================================
class User(Level):
    '''Create new rpg4go user, and store any rpg4go related user data into one
object. Object's rpg4go related data can also be saved to and loaded from user
database.'''

    # ===========
    # > Generic
    # ===========
    def __init__(self, userid, level, exp, skillPoints, skills):
        '''Constructor creates User object out of given userid and gorpg information.'''

        # Save properties for user
        self.userid = userid
        self.player = Player.PlayerOfUserid(userid)
        self.steamid = get_uniqueid_from_player(self.player)
        self.index = Engine.IndexOfEdict(self.edict)

        # Leveling attributes
        super().__init__(level)
        self.exp = exp

        # rpg4go attributes
        self.skillPoints = skillPoints
        self.skills = skills

    @property
    def edict(self):
        '''Return edict for self.player.'''
        return Player.EdictOfPlayer(self.player)

    def message(self, text):
        '''Show text on user's chat window.'''
        chat(self.index, text)
    
    @staticmethod
    def new(userid):
        '''Create new user from userid with default rpg4go values.'''

        # Create new user with default values
        user = User(userid, 0, 0, 0, [])

        # Initialize skills for user
        user.initSkills()

        # Return new user
        return user
        
    
    # ============
    # > Skills
    # ============
    def getSkillForSkillClass(self, cSkill):
        '''Return user's corresponding skill for cSkill.'''

        # Loop through all user's skills
        for skill in self.skills:

            # See if skill's class matches given cSkill
            if skill.__class__ == cSkill:

                # Return skill
                return skill

    def initSkills(self):
        '''Initialize any missing skills for user.'''
        
        # Create a list for storing user's cSkills
        userskills = []

        # Loop through all user's existing skills
        for skill in self.skills:

            # Make sure there will be no duplicate cSkills
            if skill.__class__ not in userskills:

                # Add cSkill into userskills list
                userskills.append(skill.__class__)
                
        # Loop through all skills
        for skill in rpg4go.globals.skill_list:

            # Did user not have this skill?
            if skill not in userskills:

                # Add new skill for user
                self.skills.append(skill(self))

    def activateSkills(self, event_type, *args):
        '''Activate all user's skills registered for given event.'''
        
        # Loop through all user's skills
        for skill in self.skills:

            # Is skill's event equal to given event?
            if skill.__class__.event == event_type:

                # Does user have any levels on this skill?
                if skill.level > 0:

                    # Activate skill's action
                    skill.action(*args)
        
    
    # ============
    # > Activity
    # ============
    @staticmethod
    def getUser(userid):
        '''Return active user from user_list if exists.'''

        # Loop through all active users
        for user in rpg4go.globals.user_list:

            # Does user have matching userid?
            if user.userid == userid:

                # Return user
                return user

    def activate(self):
        '''Activate new user. Raises an error on already active users.'''

        # Is user already active?
        if User.getUser(self.userid):

            # Raise an error
            raise KeyError("Attempt to activate already active user with userid '%i'" %self.userid)

        # Add user to user_list
        rpg4go.globals.addUser(self)

    def release(self):
        '''Release active user from server. Raises an error on unactive users.'''

        # If user is not active
        if not User.getUser(self.userid):

            # Raise an error
            raise KeyError("Attempt to release unactive user with userid '%i'" %self.userid)

        # Delete user from user_list
        rpg4go.globals.removeUser(self)


    # ============
    # > Database
    # ============
    @staticmethod
    def adapt(user):
        '''Adapt user's data into a string for database storage.'''
        
        # Userinfo
        text = "%i\n%i\n%i\t" %(user.level, user.exp, user.skillPoints)
        
        # Skillinfo
        for skill in user.skills:
            text += "%s:%i\n" %(skill.__class__.__name__, skill.level)
            
        # Return withouth last '\n'
        return text.strip()
    
    @staticmethod
    def convert(text):
        '''Convert user's rpg4go data from bytes to strings.'''

        # Is there anything to convert?
        if text:

            # Decode bytes to utf-8 string
            text = text.decode("utf-8")

            # Split text to [userinfo, skillinfo]
            text = text.split("\t")
            
            # Split userinfo string to [level, exp, skillPoints]
            userinfo = text[0].split("\n")

            # Convert userinfo to integers
            for i in range(0, len(userinfo)):
                userinfo[i] = int(userinfo[i])
                
            # Create a dictionary for storing skillinfo
            skillinfo = {}

            # Loop through all the skill strings [s1:lvl1, s2:lvl2, ...]
            for skill in text[1].split("\n"):

                # Split skill "name:level" into ["name", "level"]
                skill = skill.split(":")

                # Add skill to skillinfo dictionary
                skillinfo[skill[0]] = int(skill[1])

            # Return all the information
            return userinfo, skillinfo

    def saveToDatabase(self):
        '''Save user's rpg4go data to database.'''

        # Create connection for database
        conn = sqlite3.connect(rpg4go.config.database)
        cur = conn.cursor()

        # Insert user's info into database
        cur.execute(
            "INSERT OR REPLACE INTO users VALUES (?,?)", (self.steamid, self))

        # Save and close connection
        cur.close()
        conn.commit()
        conn.close()

    def loadFromDatabase(self):
        '''Load user's gorpg data from database.'''
        
        # Create connection for database
        conn = sqlite3.connect(rpg4go.config.database, detect_types=sqlite3.PARSE_DECLTYPES)
        cur = conn.cursor()

        # Retrieve user's info from database
        cur.execute(
            "SELECT * FROM users WHERE steamid=?", (self.steamid,))
        info = cur.fetchone()

        # Close connection
        cur.close()
        conn.close()
        
        # Was there data for given steamid?
        if info: #(steamid, ([level, exp, skillPoints], {s1, s2, ...}))

            # Save info into variables for simplicity
            userinfo, skillinfo = info[1][0], info[1][1]

            # Set user's rpg4go attributes
            self.level = userinfo[0]
            self.exp = userinfo[1]
            self.skillPoints = userinfo[2]
            
            # Loop through all the skills
            for skill in rpg4go.globals.skill_list:

                # Was there data for this skill? 
                if skill.__name__ in skillinfo:

                    # Remove user's old skill
                    self.skills.remove(self.getSkillForSkillClass(skill))

                    # Add new skill with skillinfo
                    self.skills.append(skill(self, skillinfo[skill.__name__]))


    # ==============
    # > Experience
    # ==============
    def addExp(self, exp):
        '''Add experience points for user.'''
        self.exp += exp

        # Get user's current level
        oldLevel = self.level

        # Did user level up (user's exp >= required exp for leveling up)
        while self.getRequiredExp() <= self.exp:

            # Remove exp required for leveling up
            self.exp -= self.getRequiredExp()

            # Add level for user
            self.level += 1

            # Add skill point for user
            self.skillPoints += 1

        # Did user level up?
        if self.level > oldLevel:

            # Message user
            self.message("You are now level %i with %i skill point(s)."
                         %(self.level, self.skillPoints))

    def removeExp(self, exp):
        '''Remove experience points from user.'''
        self.exp -= exp

        # Does user have negative exp? (level down)
        while self.exp < 0:

            # Remove level from user
            self.level -= 1

            # Did user go under level 0 already?
            if self.level < 0:

                # Reset user's exp and level
                self.exp = 0
                self.level = 0

                # Exit
                return

            # Add required experience for leveling up (down in this case)
            self.exp += self.getRequiredExp()

    def getRequiredExp(self):
        '''Get required experience points for user to level up.'''

        # Make sure user is not at negative level
        if self.level >= 0:

            # Return value using exp_formula with user's level
            return eval(rpg4go.config.exp_formula.replace("?", str(self.level)))
